# https://peps.python.org/pep-0496/
# https://peps.python.org/pep-0508/
import logging
import os
from typing import List, Union

from packaging.markers import InvalidMarker, Op, UndefinedComparison, UndefinedEnvironmentName, Variable
from pkg_resources import Requirement

LOGGER = logging.getLogger(__name__)


STANDARD = (
    'os_name',
    'sys_platform',
    'platform_machine',
    'platform_python_implementation',
    'platform_release',
    'platform_system',
    'platform_version',
    'python_version',
    'python_full_version',
    'implementation_name',
    'implementation_version',
)


def filter_requirements(requirements: List[Requirement]):
    dependencies = []
    for require in requirements:
        req = filter_requirement(require)
        if req:
            dependencies.append(str(req))
    return dependencies


def filter_requirement(require: Requirement):
    try:
        if require.marker is None:
            return require
        if require.marker.evaluate(os.environ):
            require.marker._markers = filter_custom_markers(require.marker._markers)
            return require
        return None
    except (InvalidMarker, UndefinedComparison, UndefinedEnvironmentName) as e:
        LOGGER.exception(f"Invalid dependency: [{str(require)}], {str(e)}")
        return None


def filter_custom_markers(markers: Union[tuple, str, list]):
    if isinstance(markers, list):
        _markers = [filter_custom_markers(marker) for marker in markers]
        for i, marker in enumerate(_markers):
            if marker is not None:
                continue
            if i >= 1 and isinstance(_markers[i - 1], (str, Op)):
                _markers[i - 1] = None
            if i < len(_markers) - 1 and isinstance(_markers[i + 1], (str, Op)):
                _markers[i + 1] = None
        _markers = list(filter(None, _markers))
        return _markers if len(_markers) > 0 else None
    elif isinstance(markers, str):
        return markers
    elif isinstance(markers, tuple):
        if any(isinstance(m, Variable) for m in markers):
            return markers
        else:
            return None
    else:
        # should not happen
        return markers



import os
import shutil
import subprocess
import sys
from typing import Optional

import click
from InquirerPy import inquirer
from pkg_resources import Requirement

from pip_inside.utils.pyproject import PyProject


def handle_add(name: str, group: Optional[str]):
    try:
        if os.environ.get('VIRTUAL_ENV') is None:
            proceed = inquirer.confirm(message='Not in virutal env, sure to proceed?', default=False).execute()
            if not proceed:
                return
        pyproject = PyProject.from_toml()
        require = Requirement(name)
        if not pyproject.add_dependency(require, group):
            click.secho("Skip, already installed as main dependency")
            return

        cmd = [shutil.which('python'), '-m', 'pip', 'install', str(require)]
        if subprocess.run(cmd, stderr=sys.stderr, stdout=sys.stdout).returncode == 0:
            pyproject.flush()
    except subprocess.CalledProcessError:
        pass



import os
from typing import Any, Dict, List, Optional, Union

import tomlkit
from pkg_resources import Requirement

from pip_inside import Aborted
from pip_inside.utils import markers

from .misc import norm_module


class PyProject:
    def __init__(self, path='pyproject.toml') -> None:
        self.path = path
        self._meta: tomlkit.TOMLDocument = tomlkit.TOMLDocument()
        self._dependencies: Dict[str, List[Requirement]] = {}

    @classmethod
    def from_toml(cls, path='pyproject.toml'):
        pyproject = cls(path)
        pyproject.load()
        return pyproject

    def _load_dependencies(self):
        key_main = 'project.dependencies'
        key_optionals = 'project.optional-dependencies'
        self._dependencies['main'] = [Requirement(dep) for dep in self.get(key_main, default=[])]
        for key, deps in self.get(key_optionals, default={}).items():
            self._dependencies[key] = [Requirement(dep) for dep in deps]

    def _dump_dependencies(self):
        for key, requires in self._dependencies.items():
            deps = tomlkit.array()
            for r in requires:
                deps.add_line(str(r))
            deps = deps.multiline(True)
            if key == 'main':
                self.set('project.dependencies', deps)
            else:
                self.set(f"project.optional-dependencies.{key}", deps)

    def validate(self):
        def check_exists(attr: str, msg: Optional[str] = None):
            if self.get(attr) is None:
                msg = f", {msg}" if msg else ''
                raise Aborted(f"Unsupported pyproject.toml, expecting: '{attr}' {msg}")
        def check_not_exists(attr: str, msg: Optional[str] = None):
            if self.get(attr) is not None:
                msg = f", {msg}" if msg else ''
                raise Aborted(f"Unsupported pyproject.toml, unexpected: '{attr}' {msg}")
        def check_equals(attr: str, val: Any, msg: Optional[str] = None):
            if self.get(attr) != val:
                msg = f", {msg}" if msg else ''
                raise Aborted(f"Unsupported pyproject.toml, expecting `{attr} = {val}` {msg}")

        if len(self._meta) == 0:
            return
        check_exists('project.name')
        check_not_exists('project.version', f"should be defined in {norm_module(self.get('project.name'))}/__init__.py")
        check_equals('project.dynamic', ['version'])
        check_exists('project.requires-python')
        check_exists('build-system')
        check_equals('build-system.build-backend', 'flit_core.buildapi', 'only supports `flit_core` backend')

    def load(self):
        if not os.path.exists(self.path):
            raise ValueError(f"'{self.path}' not found")

        with open(self.path, 'r') as f:
            self._meta = tomlkit.load(f)
        self.validate()
        self._load_dependencies()

    def flush(self):
        self._dump_dependencies()
        with open(self.path, "w") as f:
            tomlkit.dump(self._meta, f)

    def update(self, key: str, value: Union[str, int, float, dict, list]):
        data = self._meta
        attrs = key.split('.')
        for attr in attrs[:-1]:
            data = data.setdefault(attr, {})
        data[attrs[-1]] = value

    def get(self, key: str, *, create_if_missing: bool = False, default = None):
        data = self._meta
        attrs = key.split('.')

        for attr in attrs[:-1]:
            if create_if_missing:
                data = data.setdefault(attr, {})
            else:
                data = data.get(attr)
                if data is None:
                    return default
        return data.setdefault(attrs[-1], default) if create_if_missing else data.get(attrs[-1], default)

    def set(self, key: str, value: Union[str, int, float, dict, list], *, create_if_missing: bool = True):
        data = self._meta
        attrs = key.split('.')

        for attr in attrs[:-1]:
            if create_if_missing:
                data = data.setdefault(attr, {})
            else:
                data = data.get(attr)
                if data is None:
                    return False
        data[attrs[-1]] = value
        return True

    def add_dependency(self, require: Requirement, group: str = 'main'):
        do_add = True
        dependencies = self._dependencies.get(group)
        if dependencies:
            for i, dep in enumerate(dependencies):
                if require.key != dep.key:
                    continue
                if str(require) == str(dep):
                    return False
                do_add = False
                dependencies[i] = require

        if do_add:
            if dependencies is None:
                self._dependencies[group] = [require]
            else:
                dependencies.append(require)
        return True

    def remove_dependency(self, require: Requirement, group: str = 'main'):
        dependencies = self._dependencies.get(group)
        if not dependencies:
            return False
        deps = [dep for dep in dependencies if dep.key != require.key]
        if len(deps) == len(dependencies):
            return False
        self._dependencies[group] = deps
        return True

    def get_for_install(self, group: str = 'main'):
        deps = self._dependencies.get(group)
        if deps is None:
            return None
        return markers.filter_requirements(deps)

    def get_dependencies_for_install(self) -> Dict[str, List[str]]:
        return {
            group: markers.filter_requirements(deps)
            for group, deps in self._dependencies.items()
        }

    def get_dependencies_with_group(self) -> Dict[Requirement, str]:
        dependencies: Dict[Requirement, str] = {}
        for group, deps in self._dependencies.items():
            for dep in deps:
                dependencies[dep] = group
        return dependencies



import re

P_HAS_VERSION_SPECIFIERS = re.compile('(?:===|~=|==|!=|<=|>=|<|>)')
URL_VERSION_SPECIFIERS = 'https://peps.python.org/pep-0440/#version-specifiers'

P_KV_SEP = re.compile('\s*=\s*')


def has_ver_spec(name: str):
    return P_HAS_VERSION_SPECIFIERS.search(name) is not None


def norm_name(name: str):
    return name.lower().replace('_', '-') if name else None


def norm_module(name: str):
    return name.lower().replace('-', '_') if name else None



import contextlib
import os
import re
import sys
from importlib.util import module_from_spec, spec_from_file_location

P = re.compile(r'__version__\s*=\s*[\'\"]([a-z0-9.-]+)[\'\"]')


def get_version_from_init(filepath: str, silent: bool = False):
    try:
        module_name = f"pi_inside.version.import.{os.path.dirname(filepath)}".replace('/', '.')
        s = spec_from_file_location(module_name, filepath)
        m = module_from_spec(s)
        with open(os.devnull, 'w') as devnull:
            with contextlib.redirect_stdout(devnull):
                sys.modules[module_name] = m
                try:
                    s.loader.exec_module(m)
                finally:
                    sys.modules.pop(module_name, None)
        return m.__version__
    except ModuleNotFoundError:  # incase running `pi` outside project's venv
        text = open(filepath).read()
        m = P.search(text)
        if m is None:
            if silent:
                return None
            else:
                raise ValueError("'__version__' not defined in 'pyproject.toml'")
        return m.groups()[0]



import os
import shutil
import signal
import sys
import venv
from pathlib import Path

import pexpect

from pip_inside import Aborted
from pip_inside.utils.misc import P_KV_SEP


def handle_shell():
    if os.name != "posix":
        raise Aborted(f"Sorry, only supports *nix, : {os.name}")

    created = _create_venv()
    _spaw_new_shell(created)


def _create_venv():
    if os.path.exists('.venv'):
        return False
    name = Path(os.getcwd()).name
    venv.create('.venv', with_pip=True, prompt=name)
    return True


def _spaw_new_shell(is_1st_time: bool):
    if os.environ.get('VIRTUAL_ENV') is not None:
        return
    def resize(*args, **kwargs) -> None:
        terminal = shutil.get_terminal_size()
        p.setwinsize(terminal.lines, terminal.columns)

    shell = os.environ.get("SHELL")
    terminal = shutil.get_terminal_size()
    p = pexpect.spawn(shell, ['-i'], dimensions=(terminal.lines, terminal.columns))
    if shell.endswith('/zsh'):
        p.setecho(False)
    if (conda_env := _find_conda_env()) is not None:
        p.sendline(f"conda activate {conda_env}")
    p.sendline('source .venv/bin/activate')
    if is_1st_time:
        p.sendline('pip install -U pip')
    signal.signal(signal.SIGWINCH, resize)
    p.interact(escape_character=None)
    p.close()
    sys.exit(p.exitstatus)


def _find_conda_env():
    try:
        with open('.venv/pyvenv.cfg') as f:
            values = dict(P_KV_SEP.split(x.strip()) for x in f.readlines())
            home = values.get('home')
            if home is None or 'conda' not in home:
                return None
            return home[:-4] if home.endswith('/bin') else home
    except Exception as e:
        return None



import collections

import click
import tomlkit

from pip_inside.utils.dependencies import Dependencies, Package


def handle_freeze():
    dependencies = Dependencies().load_dependencies()
    requirements = collections.defaultdict(list)
    for child in dependencies._root.children:
        requirements[child.group].append(f"{child.name}=={child.version}")
        for group, dep in _find_installed_child(child):
            requirements[group or child.group].append(dep)
    with open('pi.lock', 'w') as f:
        tomlkit.dump(requirements, f)
    click.secho(f"Generated pi.lock", fg='bright_cyan')


def _find_installed_child(pkg: Package):
    for child in pkg.children:
        yield child.group, f"{child.name}=={child.version}"
        for group, dep in _find_installed_child(child):
            yield group or child.group, dep



import collections
import os
import shutil
import site
from pathlib import Path
from typing import Dict, List, Optional, Set

import click
from pkg_resources import Distribution, Requirement, find_distributions

from .misc import norm_name
from .pyproject import PyProject

ROOT = 'root'
DEPENDENCIES_COMMON = [
    'pip', 'certifi', 'setuptools', 'ipython', 'poetry', 'requests', 'idna', 'urllib3', 'wheel', 'tomlkit', 'pip-inside',
]
COLOR_MAIN = 'blue'
COLOR_OPTIONAL = 'green'
COLOR_SUBS = 'white'


def get_name_fg_by_group(group):
    if group is None:
        return COLOR_SUBS
    return COLOR_MAIN if group == 'main' else COLOR_OPTIONAL


def get_site_package_path():
    env_root = Path(shutil.which('python')).parent.parent
    paths = site.getsitepackages(prefixes=[env_root])
    path = paths[0]
    if os.path.exists(path):
        return path

    return next((env_root / 'lib').glob('python*')) / 'site-packages'


class Distributions:
    def __init__(self) -> None:
        self._distributions = self._find_distributions()

    def _find_distributions(self) -> Dict[str, Distribution]:
        distributions: Dict[str, Distribution] = {}
        site_package_path = get_site_package_path()
        for dist in find_distributions(site_package_path):
            distributions[dist.key] = dist
        return distributions

    def get_all(self) -> List[Distribution]:
        return list(self._distributions.values())

    def get(self, name) -> Distribution:
        return self._distributions.get(norm_name(name))


class TreeEntry:
    def __init__(self, prefix, package: 'Package') -> None:
        self.prefix = prefix
        self.package = package

    def __str__(self):
        return f"{self.prefix} {self.package}"

    def echo(self):
        name = click.style(self.package.name, fg=get_name_fg_by_group(self.package.group))
        installed = f"installed: {self.package.version}" if self.package.version else click.style('[not installed]', fg='yellow')
        click.echo(f"{self.prefix} {name} [required: {self.package.specs or '*'}, {installed}]")


class Package:

    def __init__(self,
                 name: str,
                 *,
                 specs: str = None,
                 group: str = None,
                 version: str = None,
                 parent: 'Package' = None) -> None:
        self.name = name
        self.specs = specs
        self.group = group
        self.version = version
        self.parent: 'Package' = parent
        self.children: List['Package'] = []

    def get_ref_path(self) -> List[str]:
        paths = []
        parent = self.parent
        while parent is not None and parent.name != ROOT:
            paths.append(parent.name)
            parent = parent.parent
        paths.reverse()
        return paths

    def echo(self):
        name = click.style(self.name, fg=get_name_fg_by_group(self.group))
        installed = f"installed: {self.version}" if self.version else click.style('[not installed]', fg='yellow')
        click.echo(f"{name} [required: {self.specs or '*'}, {installed}]")

    def tree_list(self, skip='│', branch='├', last='└', hyphen='─', prefix='') -> str:
        n_children = len(self.children)
        for i, child in enumerate(self.children):
            if i < n_children - 1:
                next_prefix = ''.join([prefix, skip, '   '])
                fork = branch
            else:
                next_prefix = ''.join([prefix, '    '])
                fork = last

            yield TreeEntry(prefix=f"{prefix}{fork}{hyphen}{hyphen}", package=child)
            yield from child.tree_list(skip, branch, last, hyphen, next_prefix)

    def __str__(self) -> str:
        if self.group:
            return f"{self.name} [{self.group}] [required: {self.specs or '*'}, installed: {self.version}]"
        else:
            return f"{self.name} [required: {self.specs or '*'}, installed: {self.version}]"

    def __repr__(self) -> str:
        return self.__str__()


class Dependencies:
    def __init__(self) -> None:
        self._distributions = Distributions()
        self._pyproject = PyProject.from_toml()
        self._cyclic_dendencies = []
        self._root: Package = Package(ROOT)
        self._root_non_dep: Package = Package(ROOT)
        self._direct_dependencies = {
            r.key: Package(r.key, specs=str(r.specifier), group=group, parent=self._root)
            for r, group in self._pyproject.get_dependencies_with_group().items()
        }

    def load_dependencies(self):
        self._root.children.clear()
        for child in self._direct_dependencies.values():
            self._root.children.append(child)
            self._load_children(child)
        return self

    def load_non_dependencies(self):
        if len(self._root.children) == 0:
            self.load_dependencies()

        self._root_non_dep.children.clear()
        project_name = self._pyproject.get('project.name')
        dependencies_project = self._get_all_project_dependencies()
        exclusion = set([project_name] + dependencies_project)
        parents = collections.defaultdict(set)
        for dist in self._distributions.get_all():
            name = norm_name(dist.key)
            if name in exclusion:
                continue
            child = Package(name, parent=self._root_non_dep)
            self._root_non_dep.children.append(child)
            parents[name].add(ROOT)
            self._load_children(child, exclusion, parents)
        self._root_non_dep.children = [child for child in self._root_non_dep.children if parents.get(child.name) == {ROOT}]
        return self

    def _load_children(self, pkg: Package, exclusion: Optional[Set[str]] = None, parents: Optional[Dict[str, Set[str]]] = None):
        dist = self._distributions.get(pkg.name)
        if dist is None:
            return
        pkg.version = dist.version
        for r in dist.requires():
            name, specs_r = norm_name(r.name), str(r.specifier)
            if exclusion is not None and name in exclusion:
                continue
            ref_path = pkg.get_ref_path()
            if name in ref_path:
                self._cyclic_dendencies.append(f"{' -> '.join(ref_path)} -> {pkg.name} -> {name}")
                continue

            dep = self._direct_dependencies.get(name)
            specs, group = (specs_r, None) if dep is None else (dep.specs or specs_r, dep.group)
            child = Package(name, specs=specs, group=group, parent=pkg)
            pkg.children.append(child)
            self._load_children(child, exclusion, parents)
            if parents is not None:
                parents[name].add(pkg.name)

    def _get_all_project_dependencies(self, exclusions: Optional[List[str]] = None) -> List[str]:
        def pkg_deps(pkg: Package):
            name = norm_name(pkg.name)
            if exclusions and (name in exclusions or pkg.name in exclusions):
                return
            yield norm_name(pkg.name)
            for pkg_child in pkg.children:
                yield from pkg_deps(pkg_child)
        dependencies = list()
        for child in self._root.children:
            for name in pkg_deps(child):
                dependencies.append(name)
        return dependencies

    def print_dependencies(self):
        if len(self._root.children) == 0:
            self.load_dependencies()
        self._print_dep_tree(self._root)

    def print_non_dependencies(self):
        if len(self._root_non_dep.children) == 0:
            self.load_non_dependencies()
        self._print_dep_tree(self._root_non_dep)

    def _print_dep_tree(self, root: Package):
        if self._cyclic_dendencies:
            click.secho('Cyclic dependencies:', fg='yellow')
            for path in self._cyclic_dendencies:
                click.secho(f"\t{path}", fg='yellow')
        key_main = click.style(COLOR_MAIN, fg=COLOR_MAIN)
        key_optional = click.style(COLOR_OPTIONAL, fg=COLOR_OPTIONAL)
        key_subs = click.style(COLOR_SUBS, fg=COLOR_SUBS)
        click.secho(f"Dependencies: (main: {key_main}, optional: {key_optional}, sub-dependencies: {key_subs})")
        for child in root.children:
            child.echo()
            for entry in child.tree_list():
                entry.echo()

    def get_unused_dependencies_for(self, require: Requirement) -> List[str]:
        name = require.key
        other_in_use = set(self._get_all_project_dependencies(exclusions=[name]))
        children = {norm_name(r.name) for r in self._distributions.get(name).requires()}
        return list(children - other_in_use)



import shutil
import subprocess
import sys

import click
from pkg_resources import Requirement

from pip_inside.utils.dependencies import Dependencies
from pip_inside.utils.pyproject import PyProject


def handle_remove(name: str, group):
    try:
        pyproject = PyProject.from_toml()
        require = Requirement(name)
        if pyproject.remove_dependency(require, group):
            pyproject.flush()
            deps = Dependencies().get_unused_dependencies_for(require)
            cmd = [shutil.which('python'), '-m', 'pip', 'uninstall', require.key, *deps, '-y']
            subprocess.run(cmd, stderr=sys.stderr, stdout=sys.stdout)
        else:
            click.secho(f"Package: [{require.key}] not found in group: [{group}]", fg='yellow')
    except subprocess.CalledProcessError:
        pass



"""
https://spdx.github.io/spdx-spec/v2.3/SPDX-license-list/
"""

from pathlib import Path

LICENSES = {
    'Apache-2.0': 'Apache License 2.0',
    'GPL-3.0-or-later': 'GNU General Public License v3.0 or later',
    'LGPL-3.0-or-later': 'GNU Lesser General Public License v3.0 or later',
    'MIT': 'MIT license',
    'MulanPSL-2.0': 'Mulan Permissive Software License, Version 2',
}


def get_file(license: str):
    return Path(__file__).parent / 'licenses' / f"{license}.txt"



import shutil
from typing import List

import click
from InquirerPy import inquirer

from . import Aborted, __version__


@click.group(invoke_without_command=True)
@click.option('-V', '--version', is_flag=True, default=False, help="show version of this tool")
@click.pass_context
def cli(ctx, version: bool):
    if ctx.invoked_subcommand:
        return
    if version:
        click.secho(f"pip-inside: {__version__}")
    else:
        click.secho(ctx.get_help())


@cli.command()
@click.option('-v', 'v', is_flag=True, default=False, help="verbovse")
def init(v: bool):
    """Init project in current directory"""
    try:
        from .commands.init import handle_init
        handle_init()
    except Aborted as e:
        click.secho(e, fg='yellow')
    except Exception as e:
        click.secho(e, fg='red')
        if v:
            import traceback
            click.secho(traceback.format_exc(), fg='red')


@cli.command()
@click.argument('name', required=False, type=str)
@click.option('-G', '--group', default='main', help='dependency group')
@click.option('-v', 'v', is_flag=True, default=False, help="verbovse")
def add(name, group, v: bool):
    """Add a package as project dependency"""
    try:
        from .commands.add import handle_add
        from .utils.packages import prompt_a_package
        click.secho(f"[python] {shutil.which('python')}", fg='cyan')
        if name:
            handle_add(name, group)
        else:
            name = prompt_a_package()
            while name is not None:
                handle_add(name, group)
                name = prompt_a_package(True)
    except Aborted as e:
        click.secho(e, fg='yellow')
    except Exception as e:
        click.secho(e, fg='red')
        if v:
            import traceback
            click.secho(traceback.format_exc(), fg='red')


@cli.command()
@click.argument('name', required=False, type=str)
@click.option('-G', '--group', default='main', show_default=True, help='dependency group')
@click.option('-v', 'v', is_flag=True, default=False, help="verbovse")
def remove(name, group, v: bool):
    """Remove a package from project dependencies"""
    try:
        from .commands.remove import handle_remove
        click.secho(f"[python] {shutil.which('python')}", fg='cyan')
        if name is None:
            name = inquirer.text(message="package name:").execute()
        handle_remove(name, group)
    except Aborted as e:
        click.secho(e, fg='yellow')
    except Exception as e:
        click.secho(e, fg='red')
        if v:
            import traceback
            click.secho(traceback.format_exc(), fg='red')


@cli.command()
@click.option('-G', '--groups', multiple=True, default=['main'], show_default=True, help='dependency groups')
@click.option('-A', '--all', is_flag=True, default=False, help="all groups")
@click.option('-v', 'v', is_flag=True, default=False, help="verbovse")
def install(groups: List[str], all: bool, v: bool):
    """Install project dependencies by groups"""
    try:
        from .commands.install import handle_install
        click.secho(f"[python] {shutil.which('python')}", fg='cyan')
        if len(groups) == 1:
            groups = groups[0].split(',')
        elif len(groups) == 0:
            groups = ['main']
        if all:
            groups = ['_all_']
        handle_install(groups)
    except Aborted as e:
        click.secho(e, fg='yellow')
    except Exception as e:
        click.secho(e, fg='red')
        if v:
            import traceback
            click.secho(traceback.format_exc(), fg='red')


@cli.command()
@click.option('--dist', default='dist', show_default=True, help='build target directory')
@click.option('-v', 'v', is_flag=True, default=False, help="verbovse")
def build(dist: str, v: bool):
    """Build the wheel and sdist"""
    try:
        from .commands.build import handle_build
        handle_build(dist)
    except Aborted as e:
        click.secho(e, fg='yellow')
    except Exception as e:
        click.secho(e, fg='red')
        if v:
            import traceback
            click.secho(traceback.format_exc(), fg='red')


@cli.command()
@click.option('-r', '--repository', default='pypi', show_default=True, help='target repository')
@click.option('--dist', default='dist', show_default=True, help='build target directory')
@click.option('-i', 'interactive', is_flag=True, default=False, help="interactive mode")
@click.option('-v', 'v', is_flag=True, default=False, help="verbovse")
def publish(repository: str, dist: str, interactive: bool, v: bool):
    """Publish the wheel and sdist to remote repository"""
    try:
        from .commands.publish import handle_publish
        handle_publish(repository, dist=dist, interactive=interactive)
    except Aborted as e:
        click.secho(e, fg='yellow')
    except Exception as e:
        click.secho(e, fg='red')
        if v:
            import traceback
            click.secho(traceback.format_exc(), fg='red')


@cli.command()
@click.option('-v', 'v', is_flag=True, default=False, help="verbovse")
def shell(v: bool):
    """Ensure '.venv' virtualenv, and new shell into it"""
    try:
        from .commands.shell import handle_shell
        handle_shell()
    except Aborted as e:
        click.secho(e, fg='yellow')
    except Exception as e:
        click.secho(e, fg='red')
        if v:
            import traceback
            click.secho(traceback.format_exc(), fg='red')


@cli.command()
@click.option('--unused', is_flag=True, default=False, help="only show unused dependencies")
@click.option('-v', 'v', is_flag=True, default=False, help="verbovse")
def show(unused: bool, v: bool):
    """Show dependency tree"""
    try:
        from .commands.show import handle_show
        click.secho(f"[python] {shutil.which('python')}", fg='cyan')
        handle_show(unused)
    except Aborted as e:
        click.secho(e, fg='yellow')
    except Exception as e:
        click.secho(e, fg='red')
        if v:
            import traceback
            click.secho(traceback.format_exc(), fg='red')


@cli.command()
@click.option('-v', 'v', is_flag=True, default=False, help="verbovse")
def freeze(v: bool):
    """Freeze dependencies into 'pi.lock'"""
    try:
        from .commands.freeze import handle_freeze
        click.secho(f"[python] {shutil.which('python')}", fg='cyan')
        handle_freeze()
    except Aborted as e:
        click.secho(e, fg='yellow')
    except Exception as e:
        click.secho(e, fg='red')
        if v:
            import traceback
            click.secho(traceback.format_exc(), fg='red')


@cli.command()
@click.option('-s', '--short', is_flag=True, default=False, help="show short version")
@click.option('-v', 'v', is_flag=True, default=False, help="verbovse")
def version(short: bool, v: bool):
    """Show version of current project"""
    try:
        from .commands.version import handle_version
        handle_version(short)
    except Aborted as e:
        click.secho(e, fg='yellow')
    except Exception as e:
        click.secho(e, fg='red')
        if v:
            import traceback
            click.secho(traceback.format_exc(), fg='red')


if __name__ == "__main__":
    cli()



import click

from pip_inside.utils import misc, versions
from pip_inside.utils.pyproject import PyProject


def handle_version(short: bool = False):
    pyproject = PyProject.from_toml()
    module = pyproject.get('project.name')
    filepath = f"{misc.norm_module(module)}/__init__.py"
    ver = versions.get_version_from_init(filepath)
    version = ver if short else f"{module}: {ver}"
    click.secho(version, fg='bright_cyan')



import os
import shutil
import subprocess
import sys
from typing import List

import click
import tomlkit

from pip_inside.utils.pyproject import PyProject


def handle_install(groups: List[str]):
    _install_from_pi_lock(groups) or _install_from_pyproject_toml(groups)


def _install_from_pi_lock(groups: List[str]):
    if not os.path.exists('pi.lock'):
        return

    with open('pi.lock', 'r') as f:
        data = tomlkit.load(f)

    if '_all_' in groups:
        for group, deps in data.items():
            _install_group(group, deps)
    else:
        for group in groups:
            _install_group(group, data.get(group, []))


def _install_from_pyproject_toml(groups: List[str]):
    pyproject = PyProject.from_toml()

    if '_all_' in groups:
        for group, deps in pyproject.get_dependencies_for_install().items():
            _install_group(group, deps)

    else:
        for group in groups:
            deps = pyproject.get_for_install(group)
            _install_group(group, deps)


def _install_group(group: str, dependencies: list):
    if dependencies is None or len(dependencies) == 0:
        click.secho(f"Group: {group}, nothing to install", fg='cyan')
        return

    click.secho(f"Group: {group}", fg='cyan')
    try:
        cmd = [shutil.which('python'), '-m', 'pip', 'install', *dependencies]
        subprocess.run(cmd, stderr=sys.stderr, stdout=sys.stdout)
    except subprocess.CalledProcessError:
        pass



import os
import shutil
import subprocess
from datetime import date
from pathlib import Path
from types import SimpleNamespace

import click
import tomlkit
from InquirerPy import inquirer
from InquirerPy.base.control import Choice

from pip_inside import Aborted
from pip_inside.utils import licenses, misc, packages, versions


def handle_init():
    click.echo(f"Initializing project in {os.getcwd()}")
    check_pyproject_toml()
    meta = collect_metadata()
    toml = build_toml(meta)

    click.secho(tomlkit.dumps(toml), fg='blue')
    proceed = inquirer.confirm(message="Proceed?", default=True).execute()
    if not proceed:
        return

    write_toml(toml)
    write_readme(meta)
    write_license(meta)
    write_resource('.gitignore')
    write_resource('.dockerignore')
    write_root_module_with_version(meta)


def check_pyproject_toml():
    if not os.path.exists('pyproject.toml'):
        return

    try:
        with open('pyproject.toml', 'r') as f:
            data = tomlkit.load(f)
        build_system = data.get('build_system')
        if build_system is not None:
            if build_system.get('build-backend') == 'flit_core.buildapi':
                raise Aborted(f"Skip, 'pyproject.toml' file already exists")
            else:
                raise Aborted(f"Skip, 'pyproject.toml' file already exists with unsupported 'build-system', only supports 'flit_core.buildapi'")
    except Aborted as e:
        raise e
    except Exception as e:
        raise Aborted(f"Skip, 'pyproject.toml' already exists, and failed to load: {e}")


def collect_metadata():
    defaults = get_defaults()
    name = inquirer.text(message="Name:", default=defaults.name, mandatory=True).execute()
    description = inquirer.text(message="Description:").execute()
    version = inquirer.text(message="Version:", default=defaults.version, mandatory=True).execute()
    author = inquirer.text(message="Author name:", default=defaults.author_name).execute()
    email = inquirer.text(message="Author email:", default=defaults.author_email).execute()

    license_choices = [Choice('skip')] + [Choice(value=n, name=f"{n} ({d})") for n, d in licenses.LICENSES.items()]
    license_name = inquirer.fuzzy(message="License:", choices=license_choices, vi_mode=True, wrap_lines=True).execute()
    requires_python = inquirer.text(message="requires-python:", default=defaults.requires_python, mandatory=True).execute()
    homepage = inquirer.text(message="Home page:").execute()
    dependencies = []

    if inquirer.confirm(message="Add dependencies?", default=True).execute():
        dependencies = collect_dependencies()

    return SimpleNamespace(
        name=name,
        description=description,
        version=version,
        author=author,
        email=email,
        license=license_name,
        requires_python=requires_python,
        homepage=homepage,
        dependencies=dependencies
    )


def collect_dependencies():
    dependencies = []
    name = packages.prompt_a_package()
    while name is not None:
        dependencies.append(name)
        name = packages.prompt_a_package(True)
    return dependencies


def get_defaults():
    author_name, author_email = get_default_author()
    return SimpleNamespace(
        name=os.path.basename(os.getcwd()).lower(),
        version='0.1.0',
        author_name=author_name,
        author_email=author_email,
        requires_python='>=3.8',
    )


def get_default_author():
    try:
        cmd = ['git', 'config', '-l']
        git_configs = subprocess.check_output(cmd, stderr=subprocess.STDOUT).decode()
        values = dict(s.split('=', 1) for s in git_configs.splitlines())
        return values.get('user.name'), values.get('user.email')
    except Exception:
        return None, None


def build_toml(meta):
    toml = tomlkit.document()
    project_table = tomlkit.table()
    project_table.add('name', meta.name)
    project_table.add('description', meta.description)
    if meta.author or meta.email:
        authors_list = tomlkit.array()
        author_inline = tomlkit.inline_table()
        author_inline.update({'name': meta.author or '', 'email': meta.email or ''})
        authors_list.append(author_inline)
        project_table.add('authors', authors_list)
    project_table.add('readme', 'README.md')
    license_inline = tomlkit.inline_table()
    if meta.license != 'skip':
        license_inline.update({'file': 'LICENSE'})
        project_table.add('license', license_inline)
        project_table.add('license-expression', meta.license)
    project_table.add('dynamic', tomlkit.array('["version"]'))
    project_table.add('requires-python', meta.requires_python)
    dependencies_list = tomlkit.array()
    dependencies_list.extend(meta.dependencies)
    project_table.add('dependencies', dependencies_list)
    toml.add('project', project_table)

    if meta.homepage:
        urls_table = tomlkit.table()
        urls_table.add('homepage', meta.homepage)
        project_table.add('urls', urls_table)

    build_system_table = tomlkit.table()
    build_system_table.add('requires', tomlkit.array('["flit_core>=3.8.0,<4"]'))
    build_system_table.add('build-backend', 'flit_core.buildapi')
    toml.add('build-system', build_system_table)

    return toml


def write_toml(toml):
    with open('pyproject.toml', 'w') as f:
        tomlkit.dump(toml, f)
        click.secho(f"Added 'pyproject.toml'", fg='bright_cyan')


def write_readme(meta):
    if os.path.exists('README.md'):
        return
    with open('README.md', 'w') as f:
        f.write(f"# {meta.name}\n\n {meta.description}\n")
        click.secho(f"Added 'README.md'", fg='bright_cyan')


def write_license(meta):
    if os.path.exists('LICENSE') or meta.license == 'skip':
        return
    license_file = licenses.get_file(meta.license)
    with license_file.open() as f_in, open('LICENSE', 'w') as f_out:
        year = date.today().year
        f_out.write(f_in.read().format(year=year, author=meta.author))
        click.secho(f"Added 'LICENSE'", fg='bright_cyan')


def write_resource(filename: str):
    target = Path(filename)
    if target.exists():
        return
    src = Path(__file__).parent / 'resources' / filename
    shutil.copyfile(src, target)
    click.secho(f"Added '{filename}'", fg='bright_cyan')


def write_root_module_with_version(meta):

    version_line = f"__version__ = '{meta.version}'\n"
    module_name = misc.norm_module(meta.name)
    os.makedirs(module_name, exist_ok=True)
    path = f"{module_name}/__init__.py"

    is_exists = os.path.exists(path)
    if is_exists:
        with open(path, 'a+') as f:
            f.seek(0)
            lines = f.readlines()
            add_version_line = True
            for i, line in enumerate(lines):
                m = versions.P.search(line)
                if m is None:
                    continue
                add_version_line = False
                version = m.groups()[0]
                if version == meta.version:
                    return
                lines[i] = version_line

            f.truncate(0)
            if add_version_line:
                i = _add_version_position(lines)
                lines.insert(i, version_line)
            for line in lines:
                f.write(line)
        click.secho(f"Updated '{module_name}/__init__.py'", fg='bright_cyan')

    else:
        with open(path, 'w') as f:
            f.write(version_line)
        click.secho(f"Added '{module_name}/__init__.py'", fg='bright_cyan')


def _add_version_position(lines: list):
    saw_contents = False
    for i, line in enumerate(lines):
        line = line.strip()
        if not line:
            if not saw_contents:
                continue
            else:
                return i

        saw_contents = True
        if line.startswith('#'):
            continue
        return i
    return len(lines)
